/*
 * (c) 2012, Vyacheslav Grigoryev, <armag_vvg@hotmail.ru>
 *
 * This file is part of "SG WinReg Demo".
 *
 * "SG WinReg Demo" is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * "SG WinReg Demo" is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef MISC_H_
#define MISC_H_

#include <string>
#include <windows.h>

#include "boost/type_traits/is_same.hpp"
#include "boost/static_assert.hpp"

#include "errors.h"


#define SG_ARRAYSIZE(x) (sizeof(x)/sizeof(x[0]))

namespace sg
{
namespace details
{

struct f_translator_fastcall {
	template <class T, class R>
	static void invoke(void (*f)(), T t) {
		reinterpret_cast<R (__fastcall *)(T)>(f)(t);
	}
};

struct f_translator_stdcall {
	template <class T, class R>
	static void invoke(void (*f)(), T t) {
		reinterpret_cast<R (__stdcall *)(T)>(f)(t);
	}
};

struct f_translator_cdecl {
	template <class T, class R>
	static void invoke(void (*f)(), T t) {
		reinterpret_cast<R (__cdecl *)(T)>(f)(t);
	}
};

} // end of namespace 'details'

// String accessers
//
template <class T> const T* get_cstr(const T* s)
{
	return s;
}

template <class T> const T* get_cstr(const std::basic_string<T> &s)
{
	return s.c_str();
}

template <class T> std::size_t get_cstr_len(const T* s)
{
	return std::char_traits<T>::length(s);
}

template <class T> std::size_t get_cstr_len(const std::basic_string<T>& s)
{
	return s.length();
}

// Scoped handle
//
template <class T>
class sh_check_nocheck {
public:
	sh_check_nocheck(T) {}
protected:
	~sh_check_nocheck() {}
};

template <class T>
class sh_check_win32_zerobad {
public:
	sh_check_win32_zerobad(T t)
	{
		BOOST_STATIC_ASSERT((boost::is_same<T, HANDLE>::value));
		if(t == 0) throw win_error();
	}
protected:
	~sh_check_win32_zerobad() {}
};

template <class T>
class sh_check_win32_invalbad {
public:
	sh_check_win32_invalbad(T t)
	{
		BOOST_STATIC_ASSERT((boost::is_same<T, HANDLE>::value));
		if(t == INVALID_HANDLE_VALUE) throw win_error();
	}
protected:
	~sh_check_win32_invalbad() {}
};


/*
 * The class "scoped_handle" is intended for storing handle
 * and it's destroying function. Handle can be detached.
 */
template
<
	class T,
	template <class> class check_policy = sh_check_nocheck
>
class scoped_handle : public check_policy<T>
{
	typedef void (*tmp_f_t)();
	typedef void (*tr_f_t)(tmp_f_t, T);

	T _t;
	tmp_f_t _f;
	tr_f_t _tf;

	scoped_handle(const scoped_handle&);
	scoped_handle& operator=(const scoped_handle&);
public:
	template <class R>
	scoped_handle(T t, R (__fastcall *f)(T))
	: check_policy<T>(t)
	, _t(t)
	, _f(reinterpret_cast<tmp_f_t>(f))
	, _tf(&details::f_translator_fastcall::invoke<T, R>)
	{}

	template <class R>
	scoped_handle(T t, R (__stdcall *f)(T))
	: check_policy<T>(t)
	, _t(t)
	, _f(reinterpret_cast<tmp_f_t>(f))
	, _tf(&details::f_translator_stdcall::invoke<T, R>)
	{}

	template <class R>
	scoped_handle(T t, R (__cdecl *f)(T))
	: check_policy<T>(t)
	, _t(t)
	, _f(reinterpret_cast<tmp_f_t>(f))
	, _tf(&details::f_translator_cdecl::invoke<T, R>)
	{}

	operator T() const { return _t; }
	T get() const { return _t; }

	T detach()
	{
		_tf = 0;
		T r; swap(r, _t);
		return r;
	}

	~scoped_handle()
	{
		if(_tf) _tf(_f, _t);
	}
};

} // end of namespace 'sg'


#endif /* MISC_H_ */
