/*
 * (c) 2012, Vyacheslav Grigoryev, <armag_vvg@hotmail.ru>
 *
 * This file is part of "SG WinReg Demo".
 *
 * "SG WinReg Demo" is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * "SG WinReg Demo" is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef WIN_REGISTRY_H_
#define WIN_REGISTRY_H_

#include <cassert>
#include <stdexcept>
#include <vector>
#include <list>
#include <iterator>
#include <string>
#include <algorithm>
#include <iterator>

#include "boost/smart_ptr.hpp"

#include "windows.h"

#include "errors.h"
#include "misc.h"


namespace sg
{

namespace details
{
template <class CharT> struct key_traits;

template <>
struct key_traits<char>
{
	static const char *empty_str;

	static LONG create_key(HKEY parent_key, const char *subkey, DWORD options, REGSAM sam_desired, HKEY *result)
	{
		return ::RegCreateKeyExA(parent_key, subkey, 0, 0, options, sam_desired, 0, result, 0);
	}
	static LONG open_key(HKEY key, const char *subkey, REGSAM sam_desired, HKEY *result)
	{
		return ::RegOpenKeyExA(key, subkey, 0, sam_desired, result);
	}
	static LONG query_value(HKEY key, const char *valname, DWORD *type, BYTE *data, DWORD *cbdata)
	{
		return ::RegQueryValueExA(key, valname, 0, type, data, cbdata);
	}
	static LONG set_value(HKEY key, const char *valname, DWORD type, const BYTE *data, DWORD cbdata)
	{
		return ::RegSetValueExA(key, valname, 0, type, data, cbdata);
	}
	static LONG delete_value(HKEY key, const char *valname)
	{
		return ::RegDeleteValueA(key, valname);
	}
	static LONG enum_value(HKEY key, DWORD index, char *name, DWORD& name_length, DWORD& type)
	{
		return ::RegEnumValueA(key, index, name, &name_length, 0, &type, 0, 0);
	}
	static LONG enum_key(HKEY key, DWORD index, char *name, DWORD& name_length)
	{
		return ::RegEnumKeyExA(key, index, name, &name_length, 0, 0, 0, 0);
	}
	static LONG delete_key(HKEY key, const char *keyname)
	{
		return ::RegDeleteKeyA(key, keyname);
	}
	static DWORD expand_string(const char* source, char* dest, DWORD size)
	{
		return ::ExpandEnvironmentStringsA(source, dest, size);
	}
};

template <>
struct key_traits<wchar_t>
{
	static const wchar_t *empty_str;

	static LONG create_key(HKEY parent_key, const wchar_t *subkey, DWORD options, REGSAM sam_desired, HKEY *result)
	{
		return ::RegCreateKeyExW(parent_key, subkey, 0, 0, options, sam_desired, 0, result, 0);
	}
	static LONG open_key(HKEY key, const wchar_t *subkey, REGSAM sam_desired, HKEY *result)
	{
		return ::RegOpenKeyExW(key, subkey, 0, sam_desired, result);
	}
	static LONG query_value(HKEY key, const wchar_t *valname, DWORD *type, BYTE *data, DWORD *cbdata)
	{
		return ::RegQueryValueExW(key, valname, 0, type, data, cbdata);
	}
	static LONG set_value(HKEY key, const wchar_t *valname, DWORD type, const BYTE *data, DWORD cbdata)
	{
		return ::RegSetValueExW(key, valname, 0, type, data, cbdata);
	}
	static LONG delete_value(HKEY key, const wchar_t *valname)
	{
		return ::RegDeleteValueW(key, valname);
	}
	static LONG enum_value(HKEY key, DWORD index, wchar_t *name, DWORD& name_length, DWORD& type)
	{
		return ::RegEnumValueW(key, index, name, &name_length, 0, &type, 0, 0);
	}
	static LONG enum_key(HKEY key, DWORD index, wchar_t *name, DWORD& name_length)
	{
		return ::RegEnumKeyExW(key, index, name, &name_length, 0, 0, 0, 0);
	}
	static LONG delete_key(HKEY key, const wchar_t *keyname)
	{
		return ::RegDeleteKeyW(key, keyname);
	}
	static DWORD expand_string(const wchar_t* source, wchar_t* dest, DWORD size)
	{
		return ::ExpandEnvironmentStringsW(source, dest, size);
	}
};

} // end of namespace details

template <class CharT>
class basic_reg_key
{
	HKEY m_key;
	bool m_is_key_initialized;

	static const char *m_trivial_logic_error_msg;

	LONG load_data(const CharT* name, DWORD& type_id,
		std::vector<BYTE>& buffer, DWORD guessed_size = 0) const;

	static std::pair<const CharT*, std::size_t>
	iter_accesser(const std::basic_string<CharT>& s)
	{
		return std::pair<const CharT*, std::size_t>(s.data(), s.length());
	}

	static std::pair<const CharT*, std::size_t>
	iter_accesser(const CharT* c)
	{
		return std::pair<const CharT*, std::size_t>(c, std::char_traits<CharT>::length(c));
	}

	static void append_iter(std::vector<CharT> &v, const std::pair<const CharT*, std::size_t> &s)
	{
		std::size_t cur_pos = v.size();
		v.resize(cur_pos + s.second + 1);
		std::char_traits<CharT>::copy(&v[cur_pos], s.first, s.second);
		v[cur_pos + s.second] = (CharT)0;
	}

	typedef void (basic_reg_key::*bool_type)() const;
	void this_type_does_not_support_comparisons() const {}

	template <class T>				// Prohibit meaningless comparisons
	bool operator==(const T& r);
	template <class T>
	bool operator!=(const T& r);

public:
	typedef boost::shared_ptr<std::list<std::basic_string<CharT> > > multisz_ptr_t;
	typedef boost::shared_ptr<std::vector<BYTE> > binary_ptr_t;
	typedef std::list<std::basic_string<CharT> > multisz_val_t;

	basic_reg_key() : m_is_key_initialized(false) {}
	template <class Str> basic_reg_key(HKEY parent_key,
		Str subkey_name, REGSAM access=KEY_ALL_ACCESS);
	template <class AnotherCharT, class Str>
	basic_reg_key(const basic_reg_key<AnotherCharT> &r,
		Str subkey_name, REGSAM access=KEY_ALL_ACCESS);
	basic_reg_key(const basic_reg_key &r);
	template <class AnotherCharT>
	basic_reg_key(const basic_reg_key<AnotherCharT> &r);
	~basic_reg_key() { close(); }

	template <class Str>
	void open(HKEY parent_key, Str subkey_name, REGSAM access=KEY_ALL_ACCESS);
	template <class AnotherCharT, class Str>
	inline void open(const basic_reg_key<AnotherCharT> &r, Str subkey_name, REGSAM access=KEY_ALL_ACCESS);

	void close() throw();

	basic_reg_key& operator=(const basic_reg_key &r);
	template <class AnotherCharT>
	basic_reg_key& operator=(const basic_reg_key<AnotherCharT> &r);
	inline void swap(basic_reg_key &r) throw();

	template <class Str> const DWORD get_type(Str name) const;

	// Getters
	//
	template <class Str> const DWORD get_DWORD(
		Str name, DWORD def, bool ignore_absence=false, bool ignore_wrong_type=false) const;
	template <class Str> const std::basic_string<CharT> get_REGSZ_or_EXPAND(
		Str name, const CharT* def=NULL, bool ignore_wrong_type=false, bool dont_expand=false) const;
	template <class Str> const std::wstring get_LINK(
		Str name, const wchar_t* def=NULL, bool ignore_wrong_type=false) const;
	template <class Str> multisz_ptr_t get_MULTISZ(
		Str name, bool ignore_absence=false, bool ignore_wrong_type=false) const;
	template <class Str> binary_ptr_t get_BINARY(
		Str name, bool ignore_absence=false, bool ignore_wrong_type=false) const;

	// Setters
	//
	template <class Str> void set_DWORD(Str name, DWORD value);
	template <class Str1, class Str2> void set_REGSZ_or_EXPAND(Str1 name, Str2 value, bool is_regsz=true);
	template <class Str1, class Str2> void set_LINK(Str1 name, Str2 value);
	template <class Str, class It> void set_MULTISZ(Str name, It begin, It end);
	template <class Str, class Ctr> inline void set_MULTISZ(Str name, const Ctr &c);
	template <class Str> void set_BINARY(Str name, const BYTE* begin, const BYTE* end);
	template <class Str> inline void set_BINARY(Str name, const std::vector<BYTE>& buffer);

	template <class Str> void delete_value(Str name);
	template <class Str> void delete_subkey(Str name);

	template <class Str> basic_reg_key create_subkey(
		Str name, bool is_volatile=false, REGSAM access=KEY_ALL_ACCESS) const;

	inline basic_reg_key& attach(HKEY key);
	inline const HKEY detach();

	operator bool_type() const {
		return m_is_key_initialized? &basic_reg_key::this_type_does_not_support_comparisons : 0;
	}

	class val_iterator
	{
	public:
		struct value_type : std::pair<std::basic_string<CharT>, DWORD>
		{
			operator const std::basic_string<CharT>&() const
			{
				return this->first;
			}
		};

		typedef std::input_iterator_tag iterator_category;
		typedef DWORD difference_type;
		typedef DWORD distance_type;
		typedef const value_type* pointer;
		typedef const value_type& reference;

		inline reference operator*() const;
		inline pointer operator->() const;
		inline val_iterator operator++(int);
		val_iterator& operator++();

		bool operator!=(const val_iterator &r) const
		{
			return m_index != r.m_index;
		}

		bool operator==(const val_iterator &r) const
		{
			return !operator!=(r);
		}

	private:
		val_iterator(const basic_reg_key &r, bool end);

		const basic_reg_key<CharT> &m_main_object;
		int m_index;
		std::vector<CharT> m_names_buffer;
		value_type m_last;

		LONG deref();

		friend class basic_reg_key;
	};

	val_iterator begin_vals() const;
	val_iterator end_vals() const;

	struct range_vals
	{
		typedef val_iterator iterator;
		typedef val_iterator const_iterator;

		const basic_reg_key &m_key_object;
		range_vals(const basic_reg_key &r)
			: m_key_object(r)
		{}
		val_iterator begin() const { return m_key_object.begin_vals(); }
		val_iterator end() const { return m_key_object.end_vals(); }
	};

	range_vals vals() const { return range_vals(*this); }

	class key_iterator
	{
	public:
		typedef std::input_iterator_tag iterator_category;
		typedef DWORD difference_type;
		typedef DWORD distance_type;
		typedef std::basic_string<CharT> value_type;
		typedef const value_type* pointer;
		typedef const value_type& reference;

		inline reference operator*() const;
		inline pointer operator->() const;
		inline key_iterator operator++(int);
		key_iterator& operator++();

		bool operator!=(const key_iterator &r) const
		{
			return m_index != r.m_index;
		}

		bool operator==(const key_iterator &r) const
		{
			return !operator!=(r);
		}

	private:
		key_iterator(const basic_reg_key &r, bool end);

		const basic_reg_key<CharT> &m_main_object;
		int m_index;
		std::vector<CharT> m_names_buffer;
		value_type m_last;

		LONG deref();

		friend class basic_reg_key;
	};

	key_iterator begin_keys() const;
	key_iterator end_keys() const;

	struct range_keys
	{
		typedef key_iterator iterator;
		typedef key_iterator const_iterator;

		const basic_reg_key &m_key_object;
		range_keys(const basic_reg_key &r)
			: m_key_object(r)
		{}
		key_iterator begin() const { return m_key_object.begin_keys(); }
		key_iterator end() const { return m_key_object.end_keys(); }
	};

	range_keys keys() const { return range_keys(*this); }

	template <class> friend class basic_reg_key;
};

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// Inline methods implementation

template <class CharT>
template <class Str>
basic_reg_key<CharT>::basic_reg_key(HKEY parent_key,
		Str subkey_name, REGSAM access)
	: m_is_key_initialized(false)
{
	open(parent_key, subkey_name, access);
}

template <class CharT>
template <class AnotherCharT, class Str>
basic_reg_key<CharT>::basic_reg_key(const basic_reg_key<AnotherCharT> &r,
		Str subkey_name, REGSAM access)
	: m_is_key_initialized(false)
{
	open(r.m_key, subkey_name, access);
}

template <class CharT>
template <class Str>
void basic_reg_key<CharT>::open(HKEY parent_key, Str subkey_name, REGSAM access)
{
	assert(!m_is_key_initialized);
	if(m_is_key_initialized)
		throw std::logic_error("attempt to open already opened key");

	LONG res = details::key_traits<CharT>::open_key(parent_key,
		get_cstr(subkey_name), access, &m_key);
	if(ERROR_SUCCESS != res)
		throw win_error("unable to open registry key", res);
	m_is_key_initialized = true;
}

template <class CharT>
template <class AnotherCharT, class Str>
void basic_reg_key<CharT>::open(const basic_reg_key<AnotherCharT> &r, Str subkey_name, REGSAM access)
{
	open(r.m_key, subkey_name, access);
}

template <class CharT>
void basic_reg_key<CharT>::swap(basic_reg_key &r) throw()
{
	using std::swap;
	swap(m_is_key_initialized, r.m_is_key_initialized);
	swap(m_key, r.m_key);
}

template <class CharT>
basic_reg_key<CharT>& basic_reg_key<CharT>::attach(HKEY key)
{
	close(); m_key = key; m_is_key_initialized = true;
	return *this;
}

template <class CharT>
template <class AnotherCharT>
basic_reg_key<CharT>& basic_reg_key<CharT>::operator=(const basic_reg_key<AnotherCharT> &r)
{
	basic_reg_key<CharT> tmp(r);
	swap(tmp);
	return *this;
}

template <class CharT>
typename basic_reg_key<CharT>::val_iterator::reference
basic_reg_key<CharT>::val_iterator::operator*() const
{
	assert(m_index != -1);
	return m_last;
}

template <class CharT>
typename basic_reg_key<CharT>::val_iterator::pointer
basic_reg_key<CharT>::val_iterator::operator->() const
{
	assert(m_index != -1);
	return &m_last;
}

template <class CharT>
typename basic_reg_key<CharT>::val_iterator
basic_reg_key<CharT>::val_iterator::operator++(int)
{
	val_iterator tmp(*this);
	operator++();
	return tmp;
}

template <class CharT>
typename basic_reg_key<CharT>::key_iterator::reference
basic_reg_key<CharT>::key_iterator::operator*() const
{
	assert(m_index != -1);
	return m_last;
}

template <class CharT>
typename basic_reg_key<CharT>::key_iterator::pointer
basic_reg_key<CharT>::key_iterator::operator->() const
{
	assert(m_index != -1);
	return &m_last;
}

template <class CharT>
typename basic_reg_key<CharT>::key_iterator
basic_reg_key<CharT>::key_iterator::operator++(int)
{
	key_iterator tmp(*this);
	operator++();
	return tmp;
}

//-------------------------------------
// Getters and setters

template <class CharT>
template <class Str>
const DWORD basic_reg_key<CharT>::get_type(Str name) const
{
	assert(m_is_key_initialized);
	if(!m_is_key_initialized)
		throw std::logic_error(m_trivial_logic_error_msg);

	DWORD type_id;
	LONG res = details::key_traits<CharT>::query_value(
			m_key, get_cstr(name), &type_id, NULL, NULL);
	if(ERROR_SUCCESS == res || ERROR_MORE_DATA == res)
		return type_id;
	throw win_error("unable to get value's type", res);
}

template <class CharT>
template <class Str>
const DWORD basic_reg_key<CharT>::get_DWORD(
	Str name, DWORD def, bool ignore_absence, bool ignore_wrong_type) const
{
	assert(m_is_key_initialized);
	if(!m_is_key_initialized)
		throw std::logic_error(m_trivial_logic_error_msg);

	DWORD type_id = 0;
	std::vector<BYTE> buffer;
	LONG res = load_data(get_cstr(name), type_id, buffer, sizeof(DWORD));

	if(ERROR_SUCCESS == res)
	{
		if(REG_DWORD == type_id && buffer.size() == sizeof(DWORD))
			return *reinterpret_cast<const DWORD*>(&buffer[0]);

		if(ignore_wrong_type)
			return def;

		throw type_error("registry have another value type");
	}

	if(ERROR_FILE_NOT_FOUND == res && ignore_absence)
		return def;

	throw win_error("unable to load registry value", res);
}

template <class CharT>
template <class Str>
void basic_reg_key<CharT>::set_DWORD(Str name, DWORD value)
{
	assert(m_is_key_initialized);
	if(!m_is_key_initialized)
		throw std::logic_error(m_trivial_logic_error_msg);

	LONG res = details::key_traits<CharT>::set_value(
			m_key, get_cstr(name), REG_DWORD, &value, sizeof(DWORD));

	if(ERROR_SUCCESS != res)
		throw win_error("unable to save DWORD value in the registry key", res);
}

template <class CharT>
template <class Str>
const std::basic_string<CharT> basic_reg_key<CharT>::get_REGSZ_or_EXPAND(
	Str name, const CharT* def, bool ignore_wrong_type, bool dont_expand) const
{
	assert(m_is_key_initialized);
	if(!m_is_key_initialized)
		throw std::logic_error(m_trivial_logic_error_msg);

	DWORD type_id = 0;
	std::vector<BYTE> buffer;
	LONG res = load_data(get_cstr(name), type_id, buffer);

	if(ERROR_SUCCESS == res)
	{
		if(REG_SZ == type_id || REG_EXPAND_SZ == type_id)
		{
			if(buffer.empty())
				return std::basic_string<CharT>();

			const CharT* begin = reinterpret_cast<const CharT*>(&buffer[0]);
			const CharT* end = begin + buffer.size() / sizeof(CharT);
			while(end != begin && *(end-1) == (CharT)0) --end;

			if(REG_SZ == type_id || (REG_EXPAND_SZ == type_id && dont_expand) ||
			   std::char_traits<CharT>::find(
					   begin, end - begin, (CharT)'%') == NULL)
			{
				return std::basic_string<CharT>(begin, end);
			}

			std::vector<CharT> expand_buffer(buffer.size() + buffer.size() / 2);
			while(true)
			{
				DWORD res = details::key_traits<CharT>::expand_string(
						begin, &expand_buffer[0], expand_buffer.size());
				if(res <= expand_buffer.size())
					return std::basic_string<CharT>(&expand_buffer[0], &expand_buffer[0] + res - 1);
				else if(res > 32 * 1024)
					throw std::runtime_error("unable to expand string from the registry");

				expand_buffer.resize(res);
			}
		}
		else
			if(def && ignore_wrong_type)
				return def;
			else
				throw type_error("registry have another value type");
	}

	if(ERROR_FILE_NOT_FOUND == res && def)
		return def;

	throw win_error("unable to load registry value", res);
}

// Specified value should be a NULL-terminated string in any case.
// Note: It's good to use reinterpret_cast<>(static_cast<>(...)) to guarantee,
// that get_cstr() returns right pointer's type (equal to current CharT).
template <class CharT>
template <class Str1, class Str2>
void basic_reg_key<CharT>::set_REGSZ_or_EXPAND(
	Str1 name, Str2 value, bool is_regsz)
{
	assert(m_is_key_initialized);
	if(!m_is_key_initialized)
		throw std::logic_error(m_trivial_logic_error_msg);

	LONG res = details::key_traits<CharT>::set_value(
			m_key, get_cstr(name), (is_regsz)? REG_SZ : REG_EXPAND_SZ,
			reinterpret_cast<const BYTE*>(static_cast<const CharT*>(get_cstr(value))),
			(get_cstr_len(value) + 1) * sizeof(CharT));
	if(ERROR_SUCCESS != res)
		throw win_error("unable to save text value in the registry key", res);
}

// "Link" type is always stored in the register as unicode string!
template <class CharT>
template <class Str>
const std::wstring basic_reg_key<CharT>::get_LINK(
	Str name, const wchar_t* def, bool ignore_wrong_type) const
{
	assert(m_is_key_initialized);
	if(!m_is_key_initialized)
		throw std::logic_error(m_trivial_logic_error_msg);

	DWORD type_id = 0;
	std::vector<BYTE> buffer;
	LONG res = load_data(get_cstr(name), type_id, buffer);

	if(ERROR_SUCCESS == res)
	{
		if(REG_LINK == type_id)
		{
			if(buffer.empty())
				return std::basic_string<CharT>();

			const wchar_t* begin = reinterpret_cast<const wchar_t*>(&buffer[0]);
			const wchar_t* end = begin + buffer.size() / sizeof(wchar_t);
			while(end != begin && *(end-1) == (CharT)0) --end;

			return std::wstring(begin, end);
		}
		else
			if(def && ignore_wrong_type)
				return def;
			else
				throw type_error("registry have another value type");
	}

	if(ERROR_FILE_NOT_FOUND && def)
		return def;

	throw win_error("unable to load registry value", res);
}

// Specified value should be a NULL-terminated string in any case.
template <class CharT>
template <class Str1, class Str2>
void basic_reg_key<CharT>::set_LINK(Str1 name, Str2 value)
{
	assert(m_is_key_initialized);
	if(!m_is_key_initialized)
		throw std::logic_error(m_trivial_logic_error_msg);

	LONG res = details::key_traits<CharT>::set_value(
			m_key, get_cstr(name), REG_LINK,
			reinterpret_cast<const BYTE*>(static_cast<const wchar_t*>(get_cstr(value))),
			(get_cstr_len(value) + 1) * sizeof(wchar_t));
	if(ERROR_SUCCESS != res)
		throw win_error("unable to save text value in the registry key", res);
}

template <class CharT>
template <class Str>
typename basic_reg_key<CharT>::multisz_ptr_t
basic_reg_key<CharT>::get_MULTISZ(Str name,
	bool ignore_absence, bool ignore_wrong_type) const
{
	assert(m_is_key_initialized);
	if(!m_is_key_initialized)
		throw std::logic_error(m_trivial_logic_error_msg);

	DWORD type_id = 0;
	std::vector<BYTE> buffer;
	LONG res = load_data(get_cstr(name), type_id, buffer);

	if(ERROR_SUCCESS == res)
	{
		multisz_ptr_t ret(new std::list<std::basic_string<CharT> >());

		if(REG_MULTI_SZ == type_id)
		{
			if(buffer.empty())
				return ret;

			const CharT* begin = reinterpret_cast<const CharT*>(&buffer[0]);
			const CharT* end = begin + buffer.size() / sizeof(CharT);

			for(const CharT* p = begin; p < end && *p != (CharT)0;)
			{
				const CharT* p_end = p;
				for(; p_end < end && *p_end != (CharT)0; ++p_end);
				ret->push_back(std::basic_string<CharT>(p, p_end));
				p = p_end + 1;
			}
			return ret;
		}
		else
			if(ignore_wrong_type)
				return ret;
			else
				throw type_error("registry have another value type");
	}

	if(ERROR_FILE_NOT_FOUND && ignore_absence)
		return multisz_ptr_t(new std::list<std::basic_string<CharT> >());

	throw win_error("unable to load registry value", res);
}

template <class CharT>
template <class Str, class It>
void basic_reg_key<CharT>::set_MULTISZ(Str name, It begin, It end)
{
	assert(m_is_key_initialized);
	if(!m_is_key_initialized)
		throw std::logic_error(m_trivial_logic_error_msg);

	std::vector<CharT> buffer;
	for(; begin != end; ++begin)
	{
		// Very useful language's property - making temporary objects
		// and manage its lifetime during all expression evaluation. Even
		// if *begin returns temporary object, it is guarantees to be alive
		// when it is added to buffer.
		append_iter(buffer,	iter_accesser(*begin));
	}
	buffer.push_back(0);

	LONG res = details::key_traits<CharT>::set_value(
		m_key, get_cstr(name), REG_MULTI_SZ,
		reinterpret_cast<const BYTE*>(&buffer[0]), buffer.size() * sizeof(CharT));
	if(ERROR_SUCCESS != res)
		throw win_error("unable to save multitext value in the registry key", res);
}

template <class CharT>
template <class Str, class Ctr>
void basic_reg_key<CharT>::set_MULTISZ(Str name, const Ctr &c)
{
	set_MULTISZ(name, c.begin(), c.end());
}

template <class CharT>
template <class Str>
typename basic_reg_key<CharT>::binary_ptr_t
basic_reg_key<CharT>::get_BINARY(Str name,
	bool ignore_absence, bool ignore_wrong_type) const
{
	assert(m_is_key_initialized);
	if(!m_is_key_initialized)
		throw std::logic_error(m_trivial_logic_error_msg);

	DWORD type_id = 0;
	binary_ptr_t res_buffer = binary_ptr_t(new std::vector<BYTE>());
	LONG res = load_data(get_cstr(name), type_id, *res_buffer);

	if(ERROR_SUCCESS == res)
	{
		if(REG_BINARY == type_id)
			return res_buffer;
		else if(ignore_wrong_type)
			return binary_ptr_t(new std::vector<BYTE>());

		throw type_error("registry have another value type");
	}

	if(ERROR_FILE_NOT_FOUND && ignore_absence)
		return binary_ptr_t(new std::vector<BYTE>());

	throw win_error("unable to load registry value", res);
}

template <class CharT>
template <class Str>
void basic_reg_key<CharT>::set_BINARY(Str name, const BYTE* begin, const BYTE* end)
{
	assert(m_is_key_initialized);
	if(!m_is_key_initialized)
		throw std::logic_error(m_trivial_logic_error_msg);

	LONG res = details::key_traits<CharT>::set_value(
		m_key, get_cstr(name), REG_BINARY, begin, end - begin);
	if(ERROR_SUCCESS != res)
		throw win_error("unable to save binary value in the registry key", res);
}

template <class CharT>
template <class Str>
void basic_reg_key<CharT>::set_BINARY(Str name, const std::vector<BYTE>& buffer)
{
	set_BINARY(name, &buffer[0], &buffer[0] + buffer.size());
}

template <class CharT>
template <class Str>
void basic_reg_key<CharT>::delete_value(Str name)
{
	assert(m_is_key_initialized);
	if(!m_is_key_initialized)
		throw std::logic_error(m_trivial_logic_error_msg);

	LONG res = details::key_traits<CharT>::delete_value(m_key, get_cstr(name));
	if(ERROR_SUCCESS != res)
		throw win_error("unable to delete value from registry key", res);
}

template <class CharT>
template <class Str>
void basic_reg_key<CharT>::delete_subkey(Str name)
{
	assert(m_is_key_initialized);
	if(!m_is_key_initialized)
		throw std::logic_error(m_trivial_logic_error_msg);

	try
	{
		std::vector<std::basic_string<CharT> > subkeys_names;
		basic_reg_key subkey(*this, name, KEY_ENUMERATE_SUB_KEYS|DELETE);

		std::copy(subkey.keys().begin(), subkey.keys().end(),
			std::back_inserter(subkeys_names));

		for(typename std::vector<std::basic_string<CharT> >::const_iterator it = subkeys_names.begin(), it_end = subkeys_names.end();
			it != it_end;
			++it)
		{
			subkey.delete_subkey(*it);
		}
	}
	catch(win_error &e)
	{
		if(ERROR_FILE_NOT_FOUND != e.code())
			throw;
	}

	LONG res = details::key_traits<CharT>::delete_key(m_key, get_cstr(name));
	if(ERROR_SUCCESS != res)
		throw win_error("unable to delete registry subkey", res);
}

template <class CharT>
template <class Str>
basic_reg_key<CharT> basic_reg_key<CharT>::create_subkey(
	Str name, bool is_volatile, REGSAM access) const
{
	assert(m_is_key_initialized);
	if(!m_is_key_initialized)
		throw std::logic_error(m_trivial_logic_error_msg);

	HKEY subkey_handle;
	basic_reg_key<CharT> subkey;
	LONG res = details::key_traits<CharT>::create_key(
			m_key, get_cstr(name), (is_volatile)? REG_OPTION_VOLATILE : 0,
			access, &subkey_handle);
	if(ERROR_SUCCESS != res)
		throw win_error("unable to create subkey");
	return subkey.attach(subkey_handle);
}

typedef basic_reg_key<char> reg_key;
typedef basic_reg_key<wchar_t> reg_wkey;

} // end of namespace sg


#endif /* WIN_REGISTRY_H_ */
