/*
 * (c) 2012, Vyacheslav Grigoryev, <armag_vvg@hotmail.ru>
 *
 * This file is part of "SG WinReg Demo".
 *
 * "SG WinReg Demo" is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * "SG WinReg Demo" is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>
#include <iomanip>
#include <iterator>
#include <locale>

#include "boost/iterator/filter_iterator.hpp"

#include "win-registry.h"

// An auxillary method for conversion UTF-16 / UCS-2 string
// to multibyte string.
std::string win_ws2string(const std::wstring &s)
{
	if(s.empty()) return std::string();

	int ret;
	static std::vector<char> buffer(64);

	while(true)
	{
		ret = ::WideCharToMultiByte(CP_ACP, 0x00000400 /* WC_NO_BEST_FIT_CHARS */,
				s.data(), s.size(), &buffer[0], buffer.size(), NULL, NULL);
		if(ret)
			return std::string(&buffer[0], &buffer[0] + ret);

		if(::GetLastError() == ERROR_INSUFFICIENT_BUFFER)
			buffer.resize(buffer.size() * 2);
		else
			throw sg::win_error("unable to convert string");
	}
}

std::wstring win_mb2wstring(const std::string &s)
{
	if(s.empty()) return std::wstring();

	int ret;
	static std::vector<wchar_t> buffer(64);

	while(true)
	{
		ret = ::MultiByteToWideChar(CP_ACP, 0,
				s.data(), s.size(), &buffer[0], buffer.size());
		if(ret)
			return std::wstring(&buffer[0], &buffer[0] + ret);

		if(::GetLastError() == ERROR_INSUFFICIENT_BUFFER)
			buffer.resize(buffer.size() * 2);
		else
			throw sg::win_error("unable to convert string");
	}
}

std::string tr_str(const std::string &s) { return s; }
std::string tr_str(const std::wstring &s) { return win_ws2string(s); }


// The method prints registry's values
//
template <class CharT>
void print_reg_values(std::ostream& os, const sg::basic_reg_key<CharT>& k, int tabsize)
{
	for(typename sg::basic_reg_key<CharT>::val_iterator it = k.begin_vals(), it_end = k.end_vals();
	    it != it_end;
	    ++it)
	{
		if(tabsize) os << std::setw(tabsize) << " ";
		os << tr_str(*it);

		try
		{
			if(it->second == REG_DWORD)
				os << " = 0x" << std::hex << k.get_DWORD(it->first, 0) << std::dec;
			else if(it->second == REG_SZ)
				os << " = \"" << tr_str(k.get_REGSZ_or_EXPAND(it->first)) << "\"";
			else if(it->second == REG_MULTI_SZ)
			{
				bool b = false;
				os << " = ";
				typename sg::basic_reg_key<CharT>::multisz_ptr_t lst = k.get_MULTISZ(it->first);
				for(typename sg::basic_reg_key<CharT>::multisz_val_t::const_iterator it2 = lst->begin(), it2_end = lst->end();
					it2 != it2_end;
					++it2)
				{
					if(!b) b = true;
					else os << ", ";
					os << "\"" << tr_str(*it2) << "\"";
				}
			}
			else if(it->second == REG_BINARY)
			{
				int n = 0;
				os << " = " << std::hex << std::setfill('0');
				typename sg::basic_reg_key<CharT>::binary_ptr_t bin_ptr = k.get_BINARY(it->first);
				for(typename std::vector<BYTE>::const_iterator it = bin_ptr->begin(), it_end = bin_ptr->end();
					it != it_end;
					++it)
				{
					os << "0x" << std::setw(2) << static_cast<int>(*it) << " ";
					if(n++ >= 8) {
						os << "..."; break;
					}
				}
				os << std::setfill(' ');
			}
		}
		catch(sg::type_error &e)
		{
			std::cerr << "EE: " << e.what() << std::endl;
		}
		os << std::endl;
	}
}

// The method recursively prints registry key
//
template <class CharT>
void print_reg(std::ostream& os, const sg::basic_reg_key<CharT>& k, int tabsize)
{
	print_reg_values(os, k, tabsize);

	for(typename sg::basic_reg_key<CharT>::key_iterator it = k.begin_keys(), it_end = k.end_keys();
	    it != it_end;
	    ++it)
	{
		if(tabsize) os << std::setw(tabsize) << " ";
		os << tr_str(*it) << std::endl;

		try
		{
			sg::basic_reg_key<CharT> subkey(k, *it, KEY_READ);
			print_reg(os, subkey, tabsize + 2);
		}
		catch(sg::win_error &e)
		{
			if(e.code() == ERROR_FILE_NOT_FOUND)
				std::cerr << "EE: skip problem key: " << tr_str(*it) << std::endl;
			else if(e.code() == ERROR_ACCESS_DENIED)
				std::cerr << "EE: skip access denied key: " << tr_str(*it) << std::endl;
			else
				throw;
		}
	}
}


int print_usage()
{
	std::cout <<
		"regkey-sample.exe [-u] [-f] [-rp <KEY>]|[-c <KEY>\\<path>\\<name>:<regsz_value>]|[-d <KEY>\\<path>]\n\n"
		"Options:\n"
		"  -rp - recursively print registry contents under HKCU.\n"
		"  -u - whether use unicode reg_key class or ascii one.\n"
		"  -c - create subkey and store value under it.\n"
		"  -d - recursively delete subkey or value.\n"
		"  -f - force deletion without question.\n\n"
		"Comments, questions: Vyacheslav Grigoryev <armag_vvg@hotmail.ru>"
		<< std::endl;

	return 0;
}

// The method deletes specified key or value.
// For deleting default value '\\' character must be the last in the subkey's path.
//
template <class CharT>
void delete_reg_part(std::basic_string<CharT> key_name)
{
	std::basic_string<CharT> name_part;
	std::size_t pos = key_name.rfind((CharT)'\\');
	if(std::string::npos != pos)
	{
		name_part = key_name.substr(pos + 1);
		key_name.erase(pos);
	}
	sg::basic_reg_key<CharT> k(HKEY_CURRENT_USER, key_name);
	try
	{
		k.delete_value(name_part);
	}
	catch(sg::win_error &e)
	{
		if(e.code() == ERROR_FILE_NOT_FOUND)
		{
			k.delete_subkey(name_part);
		}
		else
			throw;
	}
}

//bool start_with_c(const std::string &s)
//{
//	return s.find(".c") == 0;
//}

int main(int argc, char *argv[])
{
//	{
//		std::cout << "'.c*' classes root subkeys:" << std::endl;
//		sg::reg_key k(HKEY_CLASSES_ROOT, "", KEY_READ);
//		std::copy(
//			boost::make_filter_iterator(&start_with_c, k.keys().begin(), k.keys().end()),
//			boost::make_filter_iterator(&start_with_c, k.keys().end(), k.keys().end()),
//			std::ostream_iterator<std::string>(std::cout, "\n"));
//		std::cout << std::endl;
//	}

	if(argc <= 1)
		return print_usage();

	bool is_force_delete = false;
	bool is_unicode = false;
	std::string key_name, value_name, value;
	enum { eIdle, eRP, eCreate, eDelete } e_mode = eIdle;
	for(unsigned short ix = 1; ix < argc; ++ix)
	{
		if(strcmp("-u", argv[ix]) == 0)
		{
			is_unicode = true;
			continue;
		}
		if(strcmp("-f", argv[ix]) == 0)
		{
			is_force_delete = true;
			continue;
		}
		if(strcmp("-rp", argv[ix]) == 0 && ix + 1 < argc)
		{
			e_mode = eRP;
			key_name = argv[ix++ + 1];
			continue;
		}
		if(strcmp("-c", argv[ix]) == 0 && ix + 1 < argc)
		{
			e_mode = eCreate;
			key_name = argv[ix++ + 1];
			std::size_t pos = key_name.rfind('\\');
			if(std::string::npos != pos)
			{
				std::size_t pos2 = key_name.find(':', pos + 1);
				if(std::string::npos != pos2)
				{
					value_name = key_name.substr(pos + 1, pos2 - pos - 1);
					value = key_name.substr(pos2 + 1);
				}
				else
					value = key_name.substr(pos + 1);
				key_name.erase(pos);
			}
			else
			{
				std::cerr << "Invalid key path was specified." << std::endl;
				return 1;
			}
			continue;
		}
		if(strcmp("-d", argv[ix]) == 0 && ix + 1 < argc)
		{
			e_mode = eDelete;
			key_name = argv[ix++ + 1];
			continue;
		}
	}

	if(e_mode == eIdle)
	{
		std::cerr << "Please, select a program mode." << std::endl;
		return 1;
	}

	if(e_mode == eDelete && !is_force_delete)
	{
		std::string s;
		std::cout << "Are you sure to delete key '" << key_name << "'? ";
		std::cout.flush();
		std::cin >> s;
		const std::locale &l = std::locale();
		for(std::string::iterator it = s.begin(), it_end = s.end(); it != it_end; ++it)
			*it = std::tolower(*it, l);
		if(s != "y" && s != "yes")
			return 0;
	}

	try
	{
		if(e_mode == eRP)
		{
			if(is_unicode)
				print_reg(std::cout, sg::reg_wkey(HKEY_CURRENT_USER, win_mb2wstring(key_name)), 0);
			else
				print_reg(std::cout, sg::reg_key(HKEY_CURRENT_USER, key_name), 0);
		}
		else if(e_mode == eCreate)
		{
			if(is_unicode)
				sg::reg_wkey(HKEY_CURRENT_USER, L"")
					.create_subkey(win_mb2wstring(key_name))
					.set_REGSZ_or_EXPAND(win_mb2wstring(value_name), win_mb2wstring(value));
			else
				sg::reg_key(HKEY_CURRENT_USER, "")
					.create_subkey(key_name)
					.set_REGSZ_or_EXPAND(value_name, value);
		}
		else if(e_mode == eDelete)
		{
			if(is_unicode)
				delete_reg_part(win_mb2wstring(key_name));
			else
				delete_reg_part(key_name);
		}
	}
	catch(std::exception &e)
	{
		std::cerr << "unexpected exception: " << e.what() << std::endl;
	}
	return 0;
}
