/*
 * (c) 2012, Vyacheslav Grigoryev, <armag_vvg@hotmail.ru>
 *
 * This file is part of "SG WinReg Demo".
 *
 * "SG WinReg Demo" is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * "SG WinReg Demo" is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "win-registry.h"

namespace sg
{

namespace details
{

const char* key_traits<char>::empty_str = "";
const wchar_t* key_traits<wchar_t>::empty_str = L"";

}

template <class CharT>
const char * basic_reg_key<CharT>::m_trivial_logic_error_msg = "attempt to use non-opened registry key";

//-----------------------------------------------------------------------------
// basic_reg_key class implementation

template <class CharT>
basic_reg_key<CharT>::basic_reg_key(const basic_reg_key &r)
{
	if(!::DuplicateHandle(::GetCurrentProcess(), r.m_key,
			::GetCurrentProcess(), (void**)&m_key, 0, FALSE, DUPLICATE_SAME_ACCESS))
	{
		throw win_error("unable to copy registry key handle");
	}
	m_is_key_initialized = true;
}

template <class CharT>
template <class AnotherCharT>
basic_reg_key<CharT>::basic_reg_key(const basic_reg_key<AnotherCharT> &r)
{
	if(!::DuplicateHandle(::GetCurrentProcess(), r.m_key,
			::GetCurrentProcess(), (void**)&m_key, 0, FALSE, DUPLICATE_SAME_ACCESS))
	{
		throw win_error("unable to copy registry key handle");
	}
	m_is_key_initialized = true;
}

template <class CharT>
void basic_reg_key<CharT>::close() throw()
{
	if(m_is_key_initialized)
		SG_VERIFY(::RegCloseKey(m_key));
	m_is_key_initialized = false;
}

template <class CharT>
basic_reg_key<CharT>& basic_reg_key<CharT>::operator=(const basic_reg_key<CharT> &r)
{
	basic_reg_key<CharT> tmp(r);
	swap(tmp);
	return *this;
}

template <class CharT>
LONG basic_reg_key<CharT>::load_data(
	const CharT* name, DWORD& type_id, std::vector<BYTE>& buffer, DWORD guessed_size) const
{
	LONG res = ERROR_SUCCESS;

	if(!guessed_size)
	{
		res = details::key_traits<CharT>::query_value(
				m_key, name, &type_id, NULL, &guessed_size);
		if(ERROR_SUCCESS != res)
			return res;
	}

	buffer.resize(guessed_size);
	if(!guessed_size) return res;

	while(true)
	{
		guessed_size = buffer.size();
		res = details::key_traits<CharT>::query_value(
				m_key, name, &type_id, &buffer[0], &guessed_size);
		if(ERROR_SUCCESS == res)
		{
			buffer.resize(guessed_size);
			return res;
		}
		if(ERROR_MORE_DATA != res)
			return res;
		buffer.resize(guessed_size + ((guessed_size >= 4)? guessed_size >> 2 : 1));
	}
}

template <class CharT>
const HKEY basic_reg_key<CharT>::detach()
{
	assert(m_is_key_initialized);
	if(!m_is_key_initialized)
		throw std::logic_error("attempt to detach uninitialized key");
	m_is_key_initialized = false;
	return m_key;
}
//-----------------------------------------------------------------------------
// values iterator implementation

template <class CharT>
typename basic_reg_key<CharT>::val_iterator
basic_reg_key<CharT>::begin_vals() const
{
	return val_iterator(*this, false);
}

template <class CharT>
typename basic_reg_key<CharT>::val_iterator
basic_reg_key<CharT>::end_vals() const
{
	return val_iterator(*this, true);
}

template <class CharT>
basic_reg_key<CharT>::val_iterator::val_iterator(const basic_reg_key<CharT> &r, bool end)
	: m_main_object(r)
	, m_index(0)
{
	if(end) { m_index = -1; return; }

	switch(LONG res = deref())
	{
	case ERROR_SUCCESS: break;
	case ERROR_MORE_DATA: break;
	case ERROR_NO_MORE_ITEMS: m_index = -1; break;
	default: throw win_error("unable to start registry enumeration", res);
	}
}

template <class CharT>
typename basic_reg_key<CharT>::val_iterator&
basic_reg_key<CharT>::val_iterator::operator++()
{
	m_index++;
	switch(LONG res = deref())
	{
	case ERROR_SUCCESS: break;
	case ERROR_MORE_DATA: break;
	case ERROR_NO_MORE_ITEMS: m_index = -1; break;
	default: throw win_error("unable to continue registry enumeration", res);
	}
	return *this;
}

template <class CharT>
LONG basic_reg_key<CharT>::val_iterator::deref()
{
	if(m_names_buffer.empty())
		m_names_buffer.resize(256);

	DWORD name_size = m_names_buffer.size();
	LONG res = details::key_traits<CharT>::enum_value(
			m_main_object.m_key, m_index, &m_names_buffer[0],
			name_size, m_last.second);

	if(ERROR_SUCCESS == res || ERROR_MORE_DATA == res)
	{
		m_last.first.assign(&m_names_buffer[0], name_size);
	}
	return res;
}
//-----------------------------------------------------------------------------
// keys iterator implementation

template <class CharT>
typename basic_reg_key<CharT>::key_iterator
basic_reg_key<CharT>::begin_keys() const
{
	return key_iterator(*this, false);
}

template <class CharT>
typename basic_reg_key<CharT>::key_iterator
basic_reg_key<CharT>::end_keys() const
{
	return key_iterator(*this, true);
}

template <class CharT>
basic_reg_key<CharT>::key_iterator::key_iterator(const basic_reg_key<CharT> &r, bool end)
	: m_main_object(r)
	, m_index(0)
{
	if(end) { m_index = -1; return; }

	switch(LONG res = deref())
	{
	case ERROR_SUCCESS: break;
	case ERROR_MORE_DATA: break;
	case ERROR_NO_MORE_ITEMS: m_index = -1; break;
	default: throw win_error("unable to start registry enumeration", res);
	}
}

template <class CharT>
LONG basic_reg_key<CharT>::key_iterator::deref()
{
	if(m_names_buffer.empty())
		m_names_buffer.resize(256);

	DWORD name_size = m_names_buffer.size();
	LONG res = details::key_traits<CharT>::enum_key(
			m_main_object.m_key, m_index, &m_names_buffer[0],
			name_size);

	if(ERROR_SUCCESS == res || ERROR_MORE_DATA == res)
	{
		m_last.assign(&m_names_buffer[0], name_size);
	}
	return res;
}

template <class CharT>
typename basic_reg_key<CharT>::key_iterator&
basic_reg_key<CharT>::key_iterator::operator++()
{
	m_index++;
	switch(LONG res = deref())
	{
	case ERROR_SUCCESS: break;
	case ERROR_MORE_DATA: break;
	case ERROR_NO_MORE_ITEMS: m_index = -1; break;
	default: throw win_error("unable to continue registry enumeration", res);
	}
	return *this;
}

//-----------------------------------------------------------------------------
// Templates explicit instantiation for common cases

template class basic_reg_key<char>;
template class basic_reg_key<wchar_t>;

template basic_reg_key<char>::basic_reg_key(const basic_reg_key<wchar_t>&);
template basic_reg_key<wchar_t>::basic_reg_key(const basic_reg_key<char>&);

} // end of namespace sg
